import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

import org.apache.jena.atlas.logging.Log;

import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.query.QueryExecutionFactory;
import com.hp.hpl.jena.query.QueryFactory;
import com.hp.hpl.jena.query.QuerySolution;
import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.rdf.model.InfModel;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.reasoner.ReasonerRegistry;

/**
 * Executes a SPARQL query on a list of data files. By default, the data files are enriched with all the triples that are entailed by them 
 * under the OWL semantics using the JENA OWL reasoner.
 * @author stijn
 *
 */
public class SparqlQueryExecutor {
	 static { Log.setCmdLogging() ; }

	
	private ModelLoader loader;
	private InfModel imodel = null;
	
	public SparqlQueryExecutor(){
		loader = new ModelLoader();
		imodel = ModelFactory.createInfModel(
				ReasonerRegistry.getOWLReasoner(), 
				loader.getModel()); 
	}
	
	public void loadData(String rdfFile) {
		loader.loadData(rdfFile);
	}
	
	public void loadData(String rdfFile, ModelLoader.DataFormats format) {
		loader.loadData(rdfFile, format);
	}
	
	public void executeQuery(String queryString) throws IOException{
        Query query = QueryFactory.create(queryString);
        
		QueryExecution qexec = QueryExecutionFactory.create(query, imodel) ;
		try {
			ResultSet results = qexec.execSelect() ;
			
			while (results.hasNext()) {
				QuerySolution solution = results.next();
				System.out.println(solution.toString());
			}
		} 
		finally { 
			qexec.close() ; 
		}
	}
	
	/**
	 * Read text from the standard input, until two consecutive blank lines.
	 * 
	 * @param is The stream from which text is to be read.
	 *
	 * @return The command read from standard input
	 */
	private static String readCommand(InputStream is) throws IOException {
		BufferedReader inread = new BufferedReader(new InputStreamReader(is));
		String line = "";
		boolean emptyLine = true;
		StringBuilder xpexpr = new StringBuilder();

		/* read to inread buffer while no double empty line is met */
		do {
			emptyLine = line.length() == 0;
			line = inread.readLine();

			if (line.length() != 0) {
				xpexpr.append(line);
				xpexpr.append("\n");
			}
		}
		while (!(line.length() == 0 && emptyLine));

		return xpexpr.toString();
	}

	/**
	 * Read text from a File into a String.
	 * 
	 * @param filePath The path to the file to be read.
	 *
	 * @return The file content, as string.
	 */
	private static String readFile(String filePath) throws IOException {
		File file = new File(filePath);
		byte[] buffer = new byte[(int)file.length()];
		BufferedInputStream f = new BufferedInputStream(new FileInputStream(file));
		f.read(buffer);
		f.close();
		return new String(buffer);
	}


	/**
	 * Load RDF file as content for a query executor.
	 *  
	 * @param sqe The Query Executor to be fed with data.
	 * @param files The RDF files containing statement to be fed to the Query Executor.
	 * @param start The first element of the files array to be taken into account.
	 * @param end The last element of the files array to be taken into account.
	 */
	private static void loadModel(SparqlQueryExecutor sqe, String files[], int start, int end) {
		for (int i = start; i <= end; i++) {
			sqe.loadData(files[i]);
		}
	}
	
	public static void main(String args[]) throws IOException {
		SparqlQueryExecutor sqe = new SparqlQueryExecutor();
		
		String queryFile = args.length == 0 ? null : args[args.length -1];
		int end = args.length -1;
		int start = 0;
		boolean queryOnStdIn = true;
		String queryString;
		
		/* detect a sparql query file */
		if (queryFile != null && queryFile.lastIndexOf('.') != -1) {
			String ext = queryFile.substring(queryFile.lastIndexOf('.') + 1);
			
			if (ext.equals("sparql")) {
				end = args.length -2;
				queryOnStdIn = false;
			}
		}
		
		/* Bad parameter (i.e. no data file) */
		if (end - start < 0) {
			System.out.println("sparqltool.jar data-files [query.sparql]\n" +
					"\n\tdata-files are rdf files (recognized extensions: ttl, rdf, n3, nt)" +
					"\n\t[query.sparql] an optional query file with \"sparql\" as extension.\n" +
					"\n\tIf no query file is specified, the query is read on standard input until double-return.");
			return;
		}
		
		/* Load the data, read the query and runs the query */
		loadModel(sqe, args, start, end);
		
		if (queryOnStdIn) {
			queryString = readCommand(System.in);
		}
		else {
			queryString = readFile(queryFile);
		}
		
		sqe.executeQuery(queryString);
	}
}
